const std = @import("std");

const models = @import("./models/root.zig");
const msg_queue = @import("msg_queue/root.zig");
const dispatchers = @import("dispatchers/root.zig");
const web = @import("web/root.zig");

const log = std.log.scoped(.craftflut);

/// Craftflut gateway configuration options
pub const Config = struct {
    /// Path to Unix socket that receives update data
    unix_socket_path: []const u8,

    /// Dispatcher config
    dispatcher: dispatchers.BlockUpdate.Config,

    /// Web server config
    web: web.Config,
};

fn openUnixSocketServer(socket_path: []const u8, max_clients: u31) !std.net.Server {
    errdefer std.posix.unlink(socket_path) catch {};

    const sockfd = try std.posix.socket(
        std.posix.AF.UNIX,
        std.posix.SOCK.STREAM | std.posix.SOCK.CLOEXEC,
        0,
    );
    const stream: std.net.Stream = .{ .handle = sockfd };
    errdefer stream.close();

    const addr = try std.net.Address.initUnix(socket_path);
    const sockaddr_ptr: *const std.posix.sockaddr = @ptrCast(&addr.un);
    try std.posix.bind(sockfd, sockaddr_ptr, addr.getOsSockLen());
    try std.posix.listen(sockfd, max_clients);

    const server: std.net.Server = .{
        .listen_address = addr,
        .stream = stream,
    };

    return server;
}

/// Block update dispatch wrapper. Panics if error occurs.
fn receiver(dispatcher: *dispatchers.BlockUpdate) void {
    dispatcher.start() catch unreachable;
}

/// Starts craftflut gateway
pub fn start(allocator: std.mem.Allocator, config: *const Config) !void {
    log.info("Starting craftflut gateway", .{});

    log.info("Opening data Unix socket: {s}", .{config.unix_socket_path});
    std.posix.unlink(config.unix_socket_path) catch {};
    var server = try openUnixSocketServer(config.unix_socket_path, 1);
    defer {
        server.deinit();
        std.posix.unlink(config.unix_socket_path) catch {};
    }

    var queue = try msg_queue.MsgQueueManaged(models.BlockUpdateMsg)
        .init(allocator, config.dispatcher.capacity);
    defer queue.deinit();

    var dispatcher = dispatchers.BlockUpdate.init(
        allocator,
        &config.dispatcher,
        queue.unmanaged(),
        &server,
    );
    log.info("Starting dispatcher", .{});
    _ = try std.Thread.spawn(.{}, receiver, .{&dispatcher});

    try web.start(allocator, &config.web, queue.unmanaged());
}
